var vows = require('vows')
var assert = require('assert')
var Runtime = require('fire').Runtime
var ValidationContext = require('../index').ValidationContext
var ValidationError = require('../index').ValidationError
var path = require('path')
require('./runtimeExtensions.js')
function createRuntime(exps, testContext) {
	if(typeof(testContext) != 'object') {
		console.trace()
	}
	var runtime = new Runtime()
	runtime.moduleRequire = function(moduleName) {
			return require(moduleName)
		}
	runtime.setBaseDir(path.dirname(module.filename))
	runtime.loadModuleInstance(require('../index.js'),'fire-validations')
	exps.forEach(function(expDef){
		runtime.registerWellKnownExpressionDefinition(expDef)
	})
	runtime.load(function(err){
		if(err){
			throw err
		}
		testContext.callback(null, runtime)
	})
}

vows.describe('validations - Contexts').addBatch({
	'When I return the validationContext variable from @Validation.Validate': {
		topic: function() {
			createRuntime([
				{
					name:"testValidationContexts",
					json: {
						"@set(person)": {
							name: "Johan",
							age: 25
						},
						"@set(vc)": null,
						"@Validations.Context(person)": {
							"@set(vc)": {
								"@get(validation)": null
							}
						},
						"@get(vc)": null
					}
				}
				], this)
		},
		"when it returns": {
			topic: function(runtime) {
				var self = this
				var contextBase = {};
				contextBase._resultCallback = function(res) {
					self.callback(null, res)
				}
				contextBase._loopCallback = function() {};
				contextBase._inputExpression  = function() {};
				contextBase._variables = {};            
				contextBase._errorCallback =  function(err) {
					self.callback(err, null)
				};
				runtime._testOnly_runExpressionByName("testValidationContexts", contextBase ,null)
			},
			"the result should be an instance of ValidationContext": function(err, res) {
				assert.isNull(err)
				assert.instanceOf(res, ValidationContext)
			},
			"the path should be /": function(err, res) {
				assert.isNull(err)
				assert.equal(res.path, '/')
				assert.isTrue(res.isRoot())
			},
			"the target should be the value from the path in the hint": function(err, res) {
				assert.isNull(err)
				assert.deepEqual(res.target, {
					name: "Johan",
					age: 25
				})
			}
		}
	},
	'When I use a @Validation.Validate nested into another': {
		topic: function() {
			createRuntime([
				{
					name:"testValidationContexts",
					json: {
						"@set(person)": {
							name: "Johan",
							age: 25
						},
						"@set(person2)": {
							name: "Johan 2",
							age: 50
						},
						"@set(vc)": null,
						"@set(vc2)": null,
						"@Validations.Context(person)": {
							"@Validations.Context(person2)": {
								"@set(vc2)": {
									"@get(validation)": null
								}
							},
							"@set(vc)": {
								"@get(validation)": null
							}
						},
						"@return": {
							vc: {
								"@get(vc)": null
							},
							vc2: {
								"@get(vc2)": null
							}
						}
					}
				}], this)
		},
		"when it returns": {
			topic: function(runtime) {
				var self = this
				var contextBase = {};
				contextBase._resultCallback = function(res) {
					self.callback(null, res)
				}
				contextBase._loopCallback = function() {};
				contextBase._inputExpression  = function() {};
				contextBase._variables = {};            
				contextBase._errorCallback =  function(err) {
					self.callback(err, null)
				};
				runtime._testOnly_runExpressionByName("testValidationContexts", contextBase ,null)
			},
			"the nested validationContext should not override the outer validationContext": function(err, res) {
				assert.isNull(err)
				assert.notEqual(res.vc, res.vc2)
			},
			"the path of both contexts should be /": function(err, res) {
				assert.isNull(err)
				assert.equal(res.vc.path, '/')
				assert.isTrue(res.vc.isRoot())
				assert.equal(res.vc2.path, '/')
				assert.isTrue(res.vc2.isRoot())
			},
			"the target of both contexts should be the value from the given path": function(err, res) {
				assert.isNull(err)
				assert.deepEqual(res.vc.target, {
					name: "Johan",
					age: 25
				})
				assert.deepEqual(res.vc2.target, {
					name: "Johan 2",
					age: 50
				})
			}
		}
	}
}).export(module);

vows.describe('validations - Members').addBatch({
	'When I use a Member validator like @Validations.Property with no context': {
		topic: function(runtime) {
			
			var self = this
			createRuntime([
				{
					name:"testMissingContextMemberValidation",
					json: {
						"@Validations.Property(member)": null
					}
				}
				], {
					callback: function(err, runtime) {

						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testMissingContextMemberValidation", contextBase ,null)
					}
				})
		},
		"it should return an error because there is no context ": function(err, res) {
			assert.isNull(res)
			assert.equal(err.error, "Member validator 'Validations.Property' requires a validation context")
		}
	},
	'When I use a Member validator like @Validations.Property with no hint': {
		topic: function(runtime) {
			
			var self = this
			createRuntime([
				{
					name:"testMissingContextMemberValidation",
					json: {
						"@Validations.Property()": null
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testMissingContextMemberValidation", contextBase ,null)
					}
				})
		},
		"it should return an error because the hint is required ": function(err, res) {
			assert.isNull(res)
			assert.equal(err.error, "Member validator 'Validations.Property' requires a hint with the name of the member to validate")
		}
	}
}).export(module);

vows.describe('validations - ValidationError').addBatch({
	'When I have a ValidationError message with name': {
		topic: function() {
			return new ValidationError('/prop/subProp','My Message',"MyErrorName")
		},
		"the toString should return the name with the message and the path": function(err) {
			assert.equal(err.toString(), "MyErrorName: 'My Message' at /prop/subProp")
		},
		"path should return the given path": function(err) {
			assert.equal(err.path, "/prop/subProp")
		},
		"name should return the given name": function(err) {
			assert.equal(err.name, "MyErrorName")
		},
		"message should return the given message": function(err) {
			assert.equal(err.message, "My Message")
		}
	},
	'When I have a ValidationError message with no name': {
		topic: function() {
			return new ValidationError('/prop/subProp','My Message')
		},
		"the toString should return the message and the path": function(err) {
			assert.equal(err.toString(), "'My Message' at /prop/subProp")
		},
		"path should return the given path": function(err) {
			assert.equal(err.path, "/prop/subProp")
		},
		"name should be null": function(err) {
			assert.isUndefined(err.name)
		},
		"message should return the given message": function(err) {
			assert.equal(err.message, "My Message")
		}
	}
}).export(module);

vows.describe('validations - @Assert.NotEmpty').addBatch({
	'When I use @Validations.Property and @Validations.NotEmpty and the field is empty': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testNonEmpty",
					json: {
						"@set(person)": {
							name: null
						},
						"@Validations.Context(person)": {
							"@Validations.Property(name)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testNonEmpty", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return an array with one error for the empty field": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					path: '/name',
					message: "Can't be empty"
				}])
		}
	},
	'When I use @Validations.Property and @Assert.NotEmpty to validate two empty fields': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testNonEmpty",
					json: {
						"@set(person)": {
							firstName: null,
							lastName: null
						},
						"@Validations.Context(person)": {
							"@Validations.Property(firstName)": {
								"@Assert.NotEmpty": null
							},
							"@Validations.Property(lastName)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testNonEmpty", contextBase ,null)
					}
				})
			
		},
		"@Validations.Context should return an array with two errors, one for every empty field": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					path: '/firstName',
					message: "Can't be empty"
				},
				{
					path: '/lastName',
					message: "Can't be empty"
				}])
		}
	},
		'When I use @Validations.Property and @Assert.NotEmpty to validate two non-empty fields': {
			topic: function() {
				var self = this
				var runtime = createRuntime([
					{
						name:"testNonEmpty",
						json: {
							"@set(person)": {
								firstName: "Johan",
								lastName: "Hernandez"
							},
							"@Validations.Context(person)": {
								"@Validations.Property(firstName)": {
									"@Assert.NotEmpty": null
								},
								"@Validations.Property(lastName)": {
									"@Assert.NotEmpty": null
								}
							}
						}
					}
					], {
						callback: function(err, runtime) {
							var contextBase = {};
							contextBase._resultCallback = function(res) {
								self.callback(null, res)
							}
							contextBase._loopCallback = function() {};
							contextBase._inputExpression  = function() {};
							contextBase._variables = {};            
							contextBase._errorCallback =  function(err) {
								self.callback(err, null)
							};
							runtime._testOnly_runExpressionByName("testNonEmpty", contextBase ,null)
						}
					})
			},
			"@Validations.Context should return an empty array because there should not be errors": function(err, res) {
				assert.isNull(err)
				assert.deepEqual(res, [])
			}
		}
}).export(module);


vows.describe('validations - @Validations.Array').addBatch({
	'When I use @Validations.Array with @Assert.NotEmpty and the array is null': {
		topic: function() {
			
			var self = this
			createRuntime([
				{
					name:"testNonEmpty",
					json: {
						"@set(person)": {
							tags: null
						},
						"@Validations.Context(person)": {
							"@Validations.Array(tags)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testNonEmpty", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return zero errors": function(err, res) {
			assert.isNull(err)
			assert.isEmpty(res)
		}
	},
	'When I use @Validations.Array with @Assert.NotEmpty with all items empty': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(person)": {
							tags: ['', null]
						},
						"@Validations.Context(person)": {
							"@Validations.Array(tags)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return an array with two errors": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					path: '/tags/0',
					message: "Can't be empty"
				},
				{
					path: '/tags/1',
					message: "Can't be empty"
				}])
		}
	},
	'When I use @Validations.Array with @Assert.NotEmpty on the members of the items': {
		topic: function() {
			
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(persons)": {
							tags: [{
								name: "",
								parentId: null
							},
							{
								name: "",
								parentId: null
							},
							{
								name: "",
								parentId: null
							}]
						},
						"@Validations.Context(persons)": {
							"@Validations.Array(tags)": {
								"@Assert.NotEmpty": null,
								"@Validations.Property(name)": {
									"@Assert.NotEmpty": null
								}
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return array with errors for the missing members": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					path: '/tags/0/name',
					message: "Can't be empty"
				},
				{
					path: '/tags/1/name',
					message: "Can't be empty"
				},
				{
					path: '/tags/2/name',
					message: "Can't be empty"
				}])
		}
	}
}).export(module);


vows.describe('validations - @Assert.Email').addBatch({
	'When I use @Assert.Email in a valid email field': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(contact)": {
							email: "johan@example.com"
						},
						"@Validations.Context(contact)": {
							"@Validations.Property(email)": {
								"@Assert.Email": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return zero errors": function(err, res) {
			assert.isNull(err)
			assert.isEmpty(res)
		}
	},
	'When I use @Assert.Email in an invalid email field': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(contact)": {
							email: "dsdasdasda"
						},
						"@Validations.Context(contact)": {
							"@Validations.Property(email)": {
								"@Assert.Email": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
							var contextBase = {};
							contextBase._resultCallback = function(res) {
								self.callback(null, res)
							}
							contextBase._loopCallback = function() {};
							contextBase._inputExpression  = function() {};
							contextBase._variables = {};            
							contextBase._errorCallback =  function(err) {
								self.callback(err, null)
							};
							runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
						
					}
				})
		},
		"@Validations.Context should return one error for the field": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					path:"/email",
					message: "Not a valid Email address"
				}
				])
		}
	},
	'When I use @Assert.Email in a valid email field': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(contact)": {
							email: "johan@example.com"
						},
						"@Validations.Context(contact)": {
							"@Validations.Property(email)": {
								"@Assert.Email": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return zero errors": function(err, res) {
			assert.isNull(err)
			assert.isEmpty(res)
		}
	}
}).export(module);


vows.describe('validations - array member validator with no hint').addBatch({
	'When I use a array validator validator with no hint': {
		topic: function() {
			
			var self = this
			var runtime = createRuntime([
				{
					name:"testExp",
					json: {
						"@set(contacts)": [
							null, ""
						],
						"@Validations.Context(contacts)": {
							"@Validations.Array": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
			

			
		},
		"@Validations.Context should return the errors with paths relative to the parent target": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					"path": "/0",
					"message": "Can't be empty"
				},
				{
					"path": "/1",
					"message": "Can't be empty"
				}
				])
		}
	}
}).export(module);


vows.describe('validations - @Assert.Error').addBatch({
	'When I use @Assert.Error to raise a validation error inside a context': {
		topic: function() {
			
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(contacts)": [],
						"@Validations.Context(contacts)": {
							"@Assert.Error": "This is my custom error report"
						}
					}
				}
				], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return an array with the custom error reported": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					"path": "/",
					"message": "This is my custom error report"
				}
				])
		}
	}
}).export(module);

vows.describe('validations - @Validations.Checkpoint').addBatch({
	'When I use @Validations.Checkpoint to suspend the validation when there are errors in the context': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(person)": {
							name: null,
							age: null,
							tags: null
						},
						"@Validations.Context(person)": {
							"@Validations.Property(name)": {
								"@Assert.NotEmpty": null
							},
							"@Validations.Property(age)": {
								"@Assert.NotEmpty": null
							},
							"@Validations.Checkpoint": null,
							"@Validations.Property(tags)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}
				], {
					callback: function(err, runtime){
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
		},
		"@Validations.Context should return errors only before the checkpoint": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					"path": "/name",
					"message": "Can't be empty"
				},
				{
					"path": "/age",
					"message": "Can't be empty"
				}
				])
		}
	},
	'When I use @Validations.Checkpoint and there are no errors in the current context': {
		topic: function() {
			var self = this
			createRuntime([
				{
					name:"testExp",
					json: {
						"@set(person)": {
							name: "Johan",
							age: 25,
							tags: null
						},
						"@Validations.Context(person)": {
							"@Validations.Property(name)": {
								"@Assert.NotEmpty": null
							},
							"@Validations.Property(age)": {
								"@Assert.NotEmpty": null
							},
							"@Validations.Checkpoint": null,
							"@Validations.Property(tags)": {
								"@Assert.NotEmpty": null
							}
						}
					}
				}], {
					callback: function(err, runtime) {
						var contextBase = {};
						contextBase._resultCallback = function(res) {
							self.callback(null, res)
						}
						contextBase._loopCallback = function() {};
						contextBase._inputExpression  = function() {};
						contextBase._variables = {};            
						contextBase._errorCallback =  function(err) {
							self.callback(err, null)
						};
						runtime._testOnly_runExpressionByName("testExp", contextBase ,null)
					}
				})
			
			
			
		},
		"@Validations.Context should return all the errors after the checkpoint": function(err, res) {
			assert.isNull(err)
			assert.deepEqual(res, [
				{
					"path": "/tags",
					"message": "Can't be empty"
				}
				])
		}
	}
}).export(module);